#!/usr/bin/python3

import binascii
import json
import logging
import re
import sys
from collections import defaultdict

MASK_MAGIC_REGEX = re.compile(r'[*?!@$]')

def to_unixnano(timestamp):
    return int(timestamp) * (10**9)

# include/atheme/channels.h
CMODE_FLAG_TO_MODE = {
    0x001: 'i', # CMODE_INVITE
    0x010: 'n', # CMODE_NOEXT
    0x080: 's', # CMODE_SEC
    0x100: 't', # CMODE_TOPIC
}

# attempt to interpret certfp as a hex-encoded SHA-256 fingerprint
def validate_certfp(certfp):
    try:
        dec = binascii.unhexlify(certfp)
    except:
        return False
    return len(dec) == 32

def convert(infile):
    out = {
        'version': 1,
        'source': 'atheme',
        'users': defaultdict(dict),
        'channels': defaultdict(dict),
    }

    group_to_founders = defaultdict(list)

    channel_to_founder = defaultdict(lambda: (None, None))

    while True:
        line = infile.readline()
        if not line:
            break
        line = line.rstrip(b'\r\n')
        try:
            line = line.decode('utf-8')
        except UnicodeDecodeError:
            line = line.decode('utf-8', 'replace')
            logging.warning("line contained invalid utf8 data " + line)
        parts = line.split(' ')
        category = parts[0]

        if category == 'GACL':
            # Note: all group definitions precede channel access entries (token CA) by design, so it
            # should be safe to read this in using one pass.
            groupname = parts[1]
            user = parts[2]
            flags = parts[3]
            if 'F' in flags:
                group_to_founders[groupname].append(user)
        elif category == 'MU':
            # user account
            # MU AAAAAAAAB shivaram $1$hcspif$nCm4r3S14Me9ifsOPGuJT. user@example.com 1600134392 1600467343 +sC default
            name = parts[2]
            user = {'name': name, 'hash': parts[3], 'email': parts[4], 'registeredAt': to_unixnano(parts[5])}
            out['users'][name].update(user)
            pass
        elif category == 'MN':
            # grouped nick
            # MN shivaram slingamn 1600218831 1600467343
            username, groupednick = parts[1], parts[2]
            if username != groupednick:
                user = out['users'][username]
                user.setdefault('additionalnicks', []).append(groupednick)
        elif category == 'MDU':
            if parts[2] == 'private:usercloak':
                username = parts[1]
                out['users'][username]['vhost'] = parts[3]
        elif category == 'MCFP':
            username, certfp = parts[1], parts[2]
            if validate_certfp(certfp):
                user = out['users'][username]
                user.setdefault('certfps', []).append(certfp.lower())
        elif category == 'MC':
            # channel registration
            # MC #mychannel 1600134478 1600467343 +v 272 0 0
            # MC #NEWCHANNELTEST 1602270889 1602270974 +vg 1 0 0 jaeger4
            chname = parts[1]
            chdata = out['channels'][chname]
            # XXX just give everyone +nt, regardless of lock status; they can fix it later
            chdata.update({'name': chname, 'registeredAt': to_unixnano(parts[2])})
            if parts[8] != '':
                chdata['key'] = parts[8]
            modes = {'n', 't'}
            mlock_on, mlock_off = int(parts[5]), int(parts[6])
            for flag, mode in CMODE_FLAG_TO_MODE.items():
                if flag & mlock_on != 0:
                    modes.add(mode)
                elif flag & mlock_off != 0 and mode in modes:
                    modes.remove(mode)
            chdata['modes'] = ''.join(sorted(modes))
            chdata['limit'] = int(parts[7])
        elif category == 'MDC':
            # auxiliary data for a channel registration
            # MDC #mychannel private:topic:setter s
            # MDC #mychannel private:topic:text hi again
            # MDC #mychannel private:topic:ts 1600135864
            chname = parts[1]
            category = parts[2]
            if category == 'private:topic:text':
                out['channels'][chname]['topic'] = line.split(maxsplit=3)[3]
            elif category == 'private:topic:setter':
                out['channels'][chname]['topicSetBy'] = parts[3]
            elif category == 'private:topic:ts':
                out['channels'][chname]['topicSetAt'] = to_unixnano(parts[3])
            elif category == 'private:mlockext':
                # the channel forward mode is +L on insp/unreal, +f on charybdis
                # charybdis has a +L ("large banlist") taking no argument
                # and unreal has a +f ("flood limit") taking two colon-delimited numbers,
                # so check for an argument that starts with a #
                if parts[3].startswith('L#') or parts[3].startswith('f#'):
                    out['channels'][chname]['forward'] = parts[3][1:]
        elif category == 'CA':
            # channel access lists
            # CA #mychannel shivaram +AFORafhioqrstv 1600134478 shivaram
            chname, username, flags, set_at = parts[1], parts[2], parts[3], int(parts[4])
            chname = parts[1]
            chdata = out['channels'][chname]
            flags = parts[3]
            set_at = int(parts[4])
            if 'amode' not in chdata:
                chdata['amode'] = {}
            # see libathemecore/flags.c: +o is op, +O is autoop, etc.
            if 'F' in flags:
                # If the username starts with "!", it's actually a GroupServ group.
                if username.startswith('!'):
                    group_founders = group_to_founders.get(username)
                    if not group_founders:
                        # skip this and warn about it later
                        continue
                    # attempt to promote the first group founder to channel founder
                    username = group_founders[0]
                    # but everyone gets the +q flag
                    for founder in group_founders:
                        chdata['amode'][founder] = 'q'
                # there can only be one founder
                preexisting_founder, preexisting_set_at = channel_to_founder[chname]
                if preexisting_founder is None or set_at < preexisting_set_at:
                    chdata['founder'] = username
                    channel_to_founder[chname] = (username, set_at)
                # but multiple people can receive the 'q' amode
                chdata['amode'][username] = 'q'
                continue
            if MASK_MAGIC_REGEX.search(username):
                # ignore groups, masks, etc. for any field other than founder
                continue
            # record the first appearing successor, if necessary
            if 'S' in flags:
                if not chdata.get('successor'):
                    chdata['successor'] = username
            # finally, handle amodes
            if 'q' in flags:
                chdata['amode'][username] = 'q'
            elif 'a' in flags:
                chdata['amode'][username] = 'a'
            elif 'o' in flags or 'O' in flags:
                chdata['amode'][username] = 'o'
            elif 'h' in flags or 'H' in flags:
                chdata['amode'][username] = 'h'
            elif 'v' in flags or 'V' in flags:
                chdata['amode'][username] = 'v'
        else:
            pass

    # do some basic integrity checks
    def validate_user(name):
        if not name:
            return False
        return bool(out['users'].get(name))

    invalid_channels = []

    for chname, chdata in out['channels'].items():
        if not validate_user(chdata.get('founder')):
            if validate_user(chdata.get('successor')):
                chdata['founder'] = chdata['successor']
            else:
                invalid_channels.append(chname)

    for chname in invalid_channels:
        logging.warning("Unable to find a valid founder for channel %s, discarding it", chname)
        del out['channels'][chname]

    return out

def main():
    if len(sys.argv) != 3:
        raise Exception("Usage: atheme2json.py atheme_db output.json")
    with open(sys.argv[1], 'rb') as infile:
        output = convert(infile)
        with open(sys.argv[2], 'w') as outfile:
            json.dump(output, outfile)

if __name__ == '__main__':
    logging.basicConfig()
    sys.exit(main())



#!/usr/bin/env python3

"""
Updates the capability definitions at irc/caps/defs.go

To add a capability, add it to the CAPDEFS list below,
then run `make capdefs` from the project root.
"""

import io
import subprocess
import sys
from collections import namedtuple

CapDef = namedtuple("CapDef", ['identifier', 'name', 'url', 'standard'])

CAPDEFS = [
    CapDef(
        identifier="AccountNotify",
        name="account-notify",
        url="https://ircv3.net/specs/extensions/account-notify-3.1.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="AccountTag",
        name="account-tag",
        url="https://ircv3.net/specs/extensions/account-tag-3.2.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="AwayNotify",
        name="away-notify",
        url="https://ircv3.net/specs/extensions/away-notify-3.1.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="Batch",
        name="batch",
        url="https://ircv3.net/specs/extensions/batch-3.2.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="CapNotify",
        name="cap-notify",
        url="https://ircv3.net/specs/extensions/cap-notify-3.2.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="ChgHost",
        name="chghost",
        url="https://ircv3.net/specs/extensions/chghost-3.2.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="EchoMessage",
        name="echo-message",
        url="https://ircv3.net/specs/extensions/echo-message-3.2.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="ExtendedJoin",
        name="extended-join",
        url="https://ircv3.net/specs/extensions/extended-join-3.1.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="ExtendedMonitor",
        name="extended-monitor",
        url="https://ircv3.net/specs/extensions/extended-monitor.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="InviteNotify",
        name="invite-notify",
        url="https://ircv3.net/specs/extensions/invite-notify-3.2.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="LabeledResponse",
        name="labeled-response",
        url="https://ircv3.net/specs/extensions/labeled-response.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="Languages",
        name="draft/languages",
        url="https://gist.github.com/DanielOaks/8126122f74b26012a3de37db80e4e0c6",
        standard="proposed IRCv3",
    ),
    CapDef(
        identifier="MessageRedaction",
        name="draft/message-redaction",
        url="https://github.com/progval/ircv3-specifications/blob/redaction/extensions/message-redaction.md",
        standard="proposed IRCv3",
    ),
    CapDef(
        identifier="MessageTags",
        name="message-tags",
        url="https://ircv3.net/specs/extensions/message-tags.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="MultiPrefix",
        name="multi-prefix",
        url="https://ircv3.net/specs/extensions/multi-prefix-3.1.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="Relaymsg",
        name="draft/relaymsg",
        url="https://github.com/ircv3/ircv3-specifications/pull/417",
        standard="proposed IRCv3",
    ),
    CapDef(
        identifier="ChannelRename",
        name="draft/channel-rename",
        url="https://ircv3.net/specs/extensions/channel-rename",
        standard="draft IRCv3",
    ),
    CapDef(
        identifier="SASL",
        name="sasl",
        url="https://ircv3.net/specs/extensions/sasl-3.2.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="ServerTime",
        name="server-time",
        url="https://ircv3.net/specs/extensions/server-time-3.2.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="SetName",
        name="setname",
        url="https://ircv3.net/specs/extensions/setname.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="STS",
        name="sts",
        url="https://ircv3.net/specs/extensions/sts.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="UserhostInNames",
        name="userhost-in-names",
        url="https://ircv3.net/specs/extensions/userhost-in-names-3.2.html",
        standard="IRCv3",
    ),
    CapDef(
        identifier="ZNCSelfMessage",
        name="znc.in/self-message",
        url="https://wiki.znc.in/Query_buffers",
        standard="ZNC vendor",
    ),
    CapDef(
        identifier="EventPlayback",
        name="draft/event-playback",
        url="https://github.com/ircv3/ircv3-specifications/pull/362",
        standard="proposed IRCv3",
    ),
    CapDef(
        identifier="ZNCPlayback",
        name="znc.in/playback",
        url="https://wiki.znc.in/Playback",
        standard="ZNC vendor",
    ),
    CapDef(
        identifier="Nope",
        name="ergo.chat/nope",
        url="https://ergo.chat/nope",
        standard="Ergo vendor",
    ),
    CapDef(
        identifier="Multiline",
        name="draft/multiline",
        url="https://github.com/ircv3/ircv3-specifications/pull/398",
        standard="proposed IRCv3",
    ),
    CapDef(
        identifier="Chathistory",
        name="draft/chathistory",
        url="https://github.com/ircv3/ircv3-specifications/pull/393",
        standard="proposed IRCv3",
    ),
    CapDef(
        identifier="AccountRegistration",
        name="draft/account-registration",
        url="https://github.com/ircv3/ircv3-specifications/pull/435",
        standard="draft IRCv3",
    ),
    CapDef(
        identifier="ReadMarker",
        name="draft/read-marker",
        url="https://github.com/ircv3/ircv3-specifications/pull/489",
        standard="draft IRCv3",
    ),
    CapDef(
        identifier="Persistence",
        name="draft/persistence",
        url="https://github.com/ircv3/ircv3-specifications/pull/503",
        standard="proposed IRCv3",
    ),
    CapDef(
        identifier="Preaway",
        name="draft/pre-away",
        url="https://github.com/ircv3/ircv3-specifications/pull/514",
        standard="proposed IRCv3",
    ),
    CapDef(
        identifier="StandardReplies",
        name="standard-replies",
        url="https://github.com/ircv3/ircv3-specifications/pull/506",
        standard="IRCv3",
    ),
   CapDef(
        identifier="NoImplicitNames",
        name="draft/no-implicit-names",
        url="https://github.com/ircv3/ircv3-specifications/pull/527",
        standard="proposed IRCv3",
    ),
   CapDef(
        identifier="ExtendedISupport",
        name="draft/extended-isupport",
        url="https://github.com/ircv3/ircv3-specifications/pull/543",
        standard="proposed IRCv3",
    ),
   CapDef(
        identifier="WebPush",
        name="draft/webpush",
        url="https://github.com/ircv3/ircv3-specifications/pull/471",
        standard="proposed IRCv3",
    ),
   CapDef(
        identifier="SojuWebPush",
        name="soju.im/webpush",
        url="https://github.com/ircv3/ircv3-specifications/pull/471",
        standard="Soju/Goguma vendor",
    ),
    CapDef(
        identifier="Metadata",
        name="draft/metadata-2",
        url="https://ircv3.net/specs/extensions/metadata",
        standard="draft IRCv3",
    ),

]

def validate_defs():
    CAPDEFS.sort(key=lambda d: d.name)
    numCaps = len(CAPDEFS)
    numNames = len(set(capdef.name for capdef in CAPDEFS))
    if numCaps != numNames:
        raise Exception("defs must have unique names, but found duplicates")
    numIdentifiers = len(set(capdef.identifier for capdef in CAPDEFS))
    if numCaps != numIdentifiers:
        raise Exception("defs must have unique identifiers, but found duplicates")

def main():
    validate_defs()
    output = io.StringIO()
    print("""
package caps

/*
	WARNING: this file is autogenerated by `make capdefs`
	DO NOT EDIT MANUALLY.
*/


    """, file=output)


    numCapabs = len(CAPDEFS)
    bitsetLen = numCapabs // 32
    if numCapabs % 32 > 0:
        bitsetLen += 1
    print ("""
const (
	// number of recognized capabilities:
	numCapabs = %d
	// length of the uint32 array that represents the bitset:
	bitsetLen = %d
)
    """ % (numCapabs, bitsetLen), file=output)

    print("const (", file=output)
    for capdef in CAPDEFS:
        print("// %s is the %s capability named \"%s\":" % (capdef.identifier, capdef.standard, capdef.name), file=output)
        print("// %s" % (capdef.url,), file=output)
        print("%s Capability = iota" % (capdef.identifier,), file=output)
        print(file=output)
    print(")", file=output)

    print("// `capabilityNames[capab]` is the string name of the capability `capab`", file=output)
    print("""var ( capabilityNames = [numCapabs]string{""", file=output)
    for capdef in CAPDEFS:
        print("\"%s\"," % (capdef.name,), file=output)
    print("})", file=output)

    # run the generated code through `gofmt -s`, which will print it to stdout
    gofmt = subprocess.Popen(['gofmt', '-s'], stdin=subprocess.PIPE)
    gofmt.communicate(input=output.getvalue().encode('utf-8'))
    if gofmt.poll() != 0:
        print(output.getvalue())
        raise Exception("gofmt failed")
    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
# updatetranslations.py
#
# tl;dr this script updates our translation file with the newest, coolest strings we've added!
# it manually searches the source code, extracts strings and then updates the language files.

# Written in 2018 by Daniel Oaks <daniel@danieloaks.net>
#
# To the extent possible under law, the author(s) have dedicated all copyright
# and related and neighboring rights to this software to the public domain
# worldwide. This software is distributed without any warranty.
#
# You should have received a copy of the CC0 Public Domain Dedication along
# with this software. If not, see
# <http://creativecommons.org/publicdomain/zero/1.0/>.

"""updatetranslations.py

Usage:
    updatetranslations.py run <irc-dir> <languages-dir>
    updatetranslations.py --version
    updatetranslations.py (-h | --help)

Options:
    <irc-dir>        Oragono's irc subdirectory where the Go code is kept.
    <languages-dir>  Languages directory."""
import os
import re
import json

from docopt import docopt
import yaml

ignored_strings = [
    'none', 'saset'
]

if __name__ == '__main__':
    arguments = docopt(__doc__, version="0.1.0")

    if arguments['run']:
        # general IRC strings
        irc_strings = []

        for subdir, dirs, files in os.walk(arguments['<irc-dir>']):
            for fname in files:
                filepath = subdir + os.sep + fname
                if filepath.endswith('.go'):
                    content = open(filepath, 'r', encoding='UTF-8').read()

                    matches = re.findall(r'\.t\("((?:[^"]|\\")+)"\)', content)
                    for match in matches:
                        if match not in irc_strings:
                            irc_strings.append(match)

                    matches = re.findall(r'\.t\(\`([^\`]+)\`\)', content)
                    for match in matches:
                        if match not in irc_strings:
                            irc_strings.append(match)

        for s in ignored_strings:
            try:
                irc_strings.remove(s)
            except ValueError:
                # ignore any that don't exist
                ...

        print("irc strings:", len(irc_strings))
        with open(os.path.join(arguments['<languages-dir>'], 'example', 'irc.lang.json'), 'w') as f:
            f.write(json.dumps({k:k for k in irc_strings}, sort_keys=True, indent=2, separators=(',', ': ')))
            f.write('\n')

        for string in irc_strings:
            if 1 < string.count('%s') + string.count('%d') + string.count('%f'):
                print('   confirm:', string)

        # help entries
        help_strings = []

        for subdir, dirs, files in os.walk(arguments['<irc-dir>']):
            for fname in files:
                filepath = subdir + os.sep + fname
                if fname == 'help.go':
                    content = open(filepath, 'r', encoding='UTF-8').read()

                    matches = re.findall(r'\`([^\`]+)\`', content)
                    for match in matches:
                        if '\n' in match and match not in help_strings:
                            help_strings.append(match)

        for s in ignored_strings:
            try:
                help_strings.remove(s)
            except ValueError:
                # ignore any that don't exist
                ...

        print("help strings:", len(help_strings))
        with open(os.path.join(arguments['<languages-dir>'], 'example', 'help.lang.json'), 'w') as f:
            f.write(json.dumps({k:k for k in help_strings}, sort_keys=True, indent=2, separators=(',', ': ')))
            f.write('\n')

        for string in help_strings:
            if 1 < string.count('%s') + string.count('%d') + string.count('%f'):
                print('   confirm:', string.split('\n')[0])

        # nickserv help entries
        help_strings = []

        for subdir, dirs, files in os.walk(arguments['<irc-dir>']):
            for fname in files:
                filepath = subdir + os.sep + fname
                if fname == 'nickserv.go':
                    content = open(filepath, 'r', encoding='UTF-8').read()

                    matches = re.findall(r'\`([^\`]+)\`', content)
                    for match in matches:
                        if match not in help_strings:
                            help_strings.append(match)

        for s in ignored_strings:
            try:
                help_strings.remove(s)
            except ValueError:
                # ignore any that don't exist
                ...

        print("nickserv help strings:", len(help_strings))
        with open(os.path.join(arguments['<languages-dir>'], 'example', 'nickserv.lang.json'), 'w') as f:
            f.write(json.dumps({k:k for k in help_strings}, sort_keys=True, indent=2, separators=(',', ': ')))
            f.write('\n')

        for string in help_strings:
            if 1 < string.count('%s') + string.count('%d') + string.count('%f'):
                print('   confirm:', string)

        # chanserv help entries
        help_strings = []

        for subdir, dirs, files in os.walk(arguments['<irc-dir>']):
            for fname in files:
                filepath = subdir + os.sep + fname
                if fname == 'chanserv.go':
                    content = open(filepath, 'r', encoding='UTF-8').read()

                    matches = re.findall(r'\`([^\`]+)\`', content)
                    for match in matches:
                        if match not in help_strings:
                            help_strings.append(match)

        for s in ignored_strings:
            try:
                help_strings.remove(s)
            except ValueError:
                # ignore any that don't exist
                ...

        print("chanserv help strings:", len(help_strings))
        with open(os.path.join(arguments['<languages-dir>'], 'example', 'chanserv.lang.json'), 'w') as f:
            f.write(json.dumps({k:k for k in help_strings}, sort_keys=True, indent=2, separators=(',', ': ')))
            f.write('\n')

        for string in help_strings:
            if 1 < string.count('%s') + string.count('%d') + string.count('%f'):
                print('   confirm:', string)

        # hostserv help entries
        help_strings = []

        for subdir, dirs, files in os.walk(arguments['<irc-dir>']):
            for fname in files:
                filepath = subdir + os.sep + fname
                if fname == 'hostserv.go':
                    content = open(filepath, 'r', encoding='UTF-8').read()

                    matches = re.findall(r'\`([^\`]+)\`', content)
                    for match in matches:
                        if match not in help_strings:
                            help_strings.append(match)

        for s in ignored_strings:
            try:
                help_strings.remove(s)
            except ValueError:
                # ignore any that don't exist
                ...

        print("hostserv help strings:", len(help_strings))
        with open(os.path.join(arguments['<languages-dir>'], 'example', 'hostserv.lang.json'), 'w') as f:
            f.write(json.dumps({k:k for k in help_strings}, sort_keys=True, indent=2, separators=(',', ': ')))
            f.write('\n')

        for string in help_strings:
            if 1 < string.count('%s') + string.count('%d') + string.count('%f'):
                print('   confirm:', string)



#!/usr/bin/python3

import binascii
import json
import logging
import re
import sys
from collections import defaultdict, namedtuple

AnopeObject = namedtuple('AnopeObject', ('type', 'kv'))

MASK_MAGIC_REGEX = re.compile(r'[*?!@]')

def access_level_to_amode(level):
    # https://wiki.anope.org/index.php/2.0/Modules/cs_xop
    if level == 'QOP':
        return 'q'
    elif level == 'SOP':
        return 'a'
    elif level == 'AOP':
        return 'o'
    elif level == 'HOP':
        return 'h'
    elif level == 'VOP':
        return 'v'

    try:
        level = int(level)
    except:
        return None
    if level >= 10000:
        return 'q'
    elif level >= 9999:
        return 'a'
    elif level >= 5:
        return 'o'
    elif level >= 4:
        return 'h'
    elif level >= 3:
        return 'v'
    else:
        return None

def to_unixnano(timestamp):
    return int(timestamp) * (10**9)

def file_to_objects(infile):
    result = []
    obj = None
    while True:
        line = infile.readline()
        if not line:
            break
        line = line.rstrip(b'\r\n')
        try:
            line = line.decode('utf-8')
        except UnicodeDecodeError:
            line = line.decode('utf-8', 'replace')
            logging.warning("line contained invalid utf8 data " + line)
        pieces = line.split(' ', maxsplit=2)
        if len(pieces) == 0:
            logging.warning("skipping blank line in db")
            continue
        if pieces[0] == 'END':
            result.append(obj)
            obj = None
        elif pieces[0] == 'OBJECT':
            obj = AnopeObject(pieces[1], {})
        elif pieces[0] == 'DATA':
            obj.kv[pieces[1]] = pieces[2]
        elif pieces[0] == 'ID':
            # not sure what these do?
            continue
        else:
            raise ValueError("unknown command found in anope db", pieces[0])
    return result

ANOPE_MODENAME_TO_MODE = {
    'NOEXTERNAL': 'n',
    'TOPIC': 't',
    'INVITE': 'i',
    'NOCTCP': 'C',
    'AUDITORIUM': 'u',
    'SECRET': 's',
}

# verify that a certfp appears to be a hex-encoded SHA-256 fingerprint;
# if it's anything else, silently ignore it
def validate_certfps(certobj):
    certfps = []
    for fingerprint in certobj.split():
        try:
            dec = binascii.unhexlify(fingerprint)
        except:
            continue
        if len(dec) == 32:
            certfps.append(fingerprint)
    return certfps

def convert(infile):
    out = {
        'version': 1,
        'source': 'anope',
        'users': defaultdict(dict),
        'channels': defaultdict(dict),
    }

    objects = file_to_objects(infile)

    lastmode_channels = set()

    for obj in objects:
        if obj.type == 'NickCore':
            username = obj.kv['display']
            userdata = {'name': username, 'hash': obj.kv['pass'], 'email': obj.kv['email']}
            certobj = obj.kv.get('cert')
            if certobj:
                userdata['certfps'] = validate_certfps(certobj)
            out['users'][username] = userdata
        elif obj.type == 'NickAlias':
            username = obj.kv['nc']
            nick = obj.kv['nick']
            userdata = out['users'][username]
            if username.lower() == nick.lower():
                userdata['registeredAt'] = to_unixnano(obj.kv['time_registered'])
            else:
                if 'additionalNicks' not in userdata:
                    userdata['additionalNicks'] = []
                userdata['additionalNicks'].append(nick)
        elif obj.type == 'ChannelInfo':
            chname = obj.kv['name']
            founder = obj.kv['founder']
            chdata = {
                'name': chname,
                'founder': founder,
                'registeredAt': to_unixnano(obj.kv['time_registered']),
                'topic': obj.kv['last_topic'],
                'topicSetBy': obj.kv['last_topic_setter'],
                'topicSetAt': to_unixnano(obj.kv['last_topic_time']),
                'amode': {founder: 'q',}
            }
            # DATA last_modes INVITE KEY,hunter2 NOEXTERNAL REGISTERED TOPIC
            last_modes = obj.kv.get('last_modes')
            if last_modes:
                modes = []
                for mode_desc in last_modes.split():
                    if ',' in mode_desc:
                        mode_name, mode_value = mode_desc.split(',', maxsplit=1)
                    else:
                        mode_name, mode_value = mode_desc, None
                    if mode_name == 'KEY':
                        chdata['key'] = mode_value
                    else:
                        modes.append(ANOPE_MODENAME_TO_MODE.get(mode_name, ''))
                chdata['modes'] = ''.join(modes)
                # prevent subsequent ModeLock objects from modifying the mode list further:
                lastmode_channels.add(chname)
            out['channels'][chname] = chdata
        elif obj.type == 'ModeLock':
            if obj.kv.get('set') != '1':
                continue
            chname = obj.kv['ci']
            if chname in lastmode_channels:
                continue
            chdata = out['channels'][chname]
            modename = obj.kv['name']
            if modename == 'KEY':
                chdata['key'] = obj.kv['param']
            else:
                oragono_mode = ANOPE_MODENAME_TO_MODE.get(modename)
                if oragono_mode is not None:
                    stored_modes = chdata.get('modes', '')
                    stored_modes += oragono_mode
                    chdata['modes'] = stored_modes
        elif obj.type == 'ChanAccess':
            chname = obj.kv['ci']
            target = obj.kv['mask']
            mode = access_level_to_amode(obj.kv['data'])
            if mode is None:
                continue
            if MASK_MAGIC_REGEX.search(target):
                continue
            chdata = out['channels'][chname]
            amode = chdata.setdefault('amode', {})
            amode[target] = mode
            chdata['amode'] = amode

    # do some basic integrity checks
    for chname, chdata in out['channels'].items():
        founder = chdata.get('founder')
        if founder not in out['users']:
            raise ValueError("no user corresponding to channel founder", chname, chdata.get('founder'))

    return out

def main():
    if len(sys.argv) != 3:
        raise Exception("Usage: anope2json.py anope.db output.json")
    with open(sys.argv[1], 'rb') as infile:
        output = convert(infile)
        with open(sys.argv[2], 'w') as outfile:
            json.dump(output, outfile)

if __name__ == '__main__':
    logging.basicConfig()
    sys.exit(main())
